from torchperf.recipe import scaling_study, comparison_study
from torchperf import perf
import torch


def test_scaling_study():
    cache = []

    @perf(o=cache, n=10)
    def mul(x):
        return x * x

    sizes = list(range(3, 6))
    inputs = (torch.rand(sz, sz, device="cuda") for sz in sizes)
    avg, std = scaling_study(mul, cache, inputs)
    assert len(avg) == len(std) == len(sizes)
    assert sum(avg) > 0 and sum(std) > 0


def test_comparison_study():
    cache = []

    @perf(o=cache, n=10)
    def mul(x):
        return x * x

    @perf(o=cache, n=10)
    def add(x):
        return x + x

    # note: add and mul share the same `rst` as output cache.

    sizes = list(range(3, 6))
    funcs = [mul, add]
    inputs = (torch.rand(sz, sz, device="cuda") for sz in sizes)
    rsts = comparison_study(funcs, cache, inputs)
    [(mul_avg, mul_std), (add_avg, add_std)] = rsts

    assert len(rsts) == len(funcs)
    assert len(mul_avg) == len(mul_std) == len(sizes)
    assert len(add_avg) == len(add_std) == len(sizes)
    assert sum(mul_avg) > 0 and sum(mul_std) > 0
    assert sum(add_avg) > 0 and sum(add_std) > 0



import sys
from subprocess import Popen, PIPE, check_output
import subprocess
import torch


def get_os_name():
    p = Popen(["lsb_release", "-d"], stdout=PIPE)
    rst = check_output(["cut", "-d:", "-f2"], stdin=p.stdout)
    return rst.decode("utf-8").strip()


def get_gpu_driver():
    p = Popen(
        ["nvidia-smi", "--query-gpu=driver_version", "--format=csv", "-i", "0"],
        stdout=PIPE,
    )
    rst = check_output(["tail", "-n1"], stdin=p.stdout)
    return rst.decode("utf-8").strip()


def get_linux_kernel():
    return subprocess.check_output(["uname", "-r"]).decode("utf-8").strip()


def show():
    def ver2str(x):
        return ".".join(map(str, x))

    print("python     ver|", ver2str(sys.version_info[:3]))
    print("pytorch    ver|", torch.__version__)
    print("CUDA       ver|", torch.version.cuda)
    print("cuDNN      ver|", torch.backends.cudnn.version() / 1000)
    print("nccl       ver|", ver2str(torch.cuda.nccl.version()))
    print("GPU Driver ver|", get_gpu_driver())
    print("Kernel     ver|", get_linux_kernel())
    print("Device Name   |", torch.cuda.get_device_name(0))
    print("OS name       |", get_os_name())
    print("num GPUs      |", torch.cuda.device_count())



import torch
from functools import wraps, partial


def perf(_func=None, *, o=None, n=1, w=1):
    r"""performance measurement decorator

    o : List, optional [=None]
        "output" elapsed times per run are stored in the list.
        If `o` is `None`, an averaged result is printed.
    n : int, optional [=1]
        "number of runs"
    w : int, optional [=1]
        "warmup", number of warmup runs.
    """

    if not _func:
        return partial(perf, o=o, n=n, w=w)

    @wraps(_func)
    def measure_time(*args, **kwargs):
        # warmup
        for _ in range(w):
            rst = _func(*args, **kwargs)

        # recording
        tot, cnt = 0.0, 0
        stt = torch.cuda.Event(enable_timing=True)
        end = torch.cuda.Event(enable_timing=True)
        for _ in range(n):
            stt.record()
            rst = _func(*args, **kwargs)
            end.record()
            torch.cuda.synchronize()
            tim = stt.elapsed_time(end)
            if o is None:
                tot += tim
                cnt += 1
            else:
                o.append(tim)
        if o is None:
            print(tot / cnt)
        return rst

    return measure_time



from typing import Callable, Iterable, Tuple, List
import numpy as np


class Stats:
    def __init__(self):
        self.avg = []
        self.std = []

    def update(self, output):
        tmp = np.asarray(output)
        self.avg.append(tmp.mean())
        self.std.append(tmp.std())


def scaling_study(func: Callable, cache: List, inputs: Iterable) -> Tuple:
    r"""study how a func performs regarding different size of inputs

    func : Callable
        a function under study that has been decorated by `perf`.
        The function can only take one argument. Consider using `partial` or a
        wrapper function to work around this limitation.
    cache : List
        the same list object passed as `o` in the decorator.
    inputs : Iterable
        an iterable of inputs. Sometimes inputs can be of large sizes. One
        should consider using a generator, e.g.
        `(torch.rand(1<<x,1<<x) for x in range(10, 15))`.
    return : Tuple
        averages and standard deviations. Both lists have the same length
        as the one of inputs.
    """

    rst = Stats()
    for input in inputs:
        cache.clear()
        _ = func(input)
        rst.update(cache)
    return rst.avg, rst.std


def comparison_study(
    funcs: List[Callable], cache: List, inputs: Iterable
) -> List[Tuple]:
    r"""study how different functions scales regarding different size of inputs
    funcs : List[Callable]
        a list of functions under study that has been decorated by `perf`.
        The function can only take one argument. Consider using `partial` or a
        wrapper function to work around this limitation. All the functions
        should share the same `cache`.
    cache : List
        the same list object passed as `o` in the decorator.
    inputs : Iterable
        an iterable of inputs. Sometimes inputs can be of large sizes. One
        should consider using a generator, e.g.
        `(torch.rand(1<<x,1<<x) for x in range(10, 15))`.
    return : List[Tuple]
        a list of the same size as the one of `func`. Each element in the list
        is a tuple of two lists: averages and standard deviations, which have
        the same length as the one of inputs.
    """

    rsts = [Stats() for _ in range(len(funcs))]

    # note: the order of the following nested for-loop cannot be reverted as
    # `inputs` could be a generator
    for input in inputs:
        for i, func in enumerate(funcs):
            cache.clear()
            _ = func(input)
            rsts[i].update(cache)

    rsts = [(stat.avg, stat.std) for stat in rsts]
    return rsts



import re
from pathlib import Path
from setuptools import find_namespace_packages, setup


def find_version(file_path: str) -> str:
    with open(file_path, "r") as fp:
        version_file = fp.read()
    ptn = r"^__version__ = ['\"]([^'\"]*)['\"]"
    version_match = re.search(ptn, version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


with open("README.md", "r") as fh:
    LONG_DESC = fh.read()
    setup(
        name="pytorch-perf",
        version=find_version(Path("./torchperf/__init__.py")),
        author="Yihui Ren",
        author_email="yren@bnl.gov",
        description="A pytorch perf decorator",
        license="BSD-3",
        long_description=LONG_DESC,
        long_description_content_type="text/markdown",
        url="https://github.com/yhren/torchperf",
        keywords="decorator cuda performance perf pytorch",
        packages=find_namespace_packages(include=["torchperf", "torchperf.*"]),
        include_package_data=True,
        python_requires=">=3.7",
        classifiers=[
            "License :: OSI Approved :: BSD License",
            "Development Status :: 2 - Pre-Alpha",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Operating System :: POSIX :: Linux",
        ],
        install_requires=["numpy"],
        # Install development dependencies with
        # pip install -r requirements/dev.txt -e .
    )



from torchperf import perf
import torch

x = torch.rand(1000, 1000, device="cuda")
y = torch.rand(1000, 1000, device="cuda")


def test_device():
    assert torch.cuda.is_available()


def test_base():
    @perf
    def mul(x, y):
        return x * y

    _ = mul(x, y)


def test_output():
    rst = []

    @perf(o=rst)
    def mul(x, y):
        return x * y

    _ = mul(x, y)
    assert len(rst) > 0


def test_repeat():
    @perf(n=10)
    def mul(x, y):
        return x * y

    _ = mul(x, y)


def test_output_repeat():
    rst = []

    @perf(o=rst, n=10)
    def mul(x, y):
        return x * y

    _ = mul(x, y)
    assert len(rst) > 0


def test_output_repeat_warmup():
    rst = []

    @perf(o=rst, n=10, w=3)
    def mul(x, y):
        return x * y

    _ = mul(x, y)
    assert len(rst) > 0
